/*
	Simple threaded portscanner by Ralph, molnar_rcs@yahoo.com
	v 1.1 / 21 jan 2006
	Can be used and distributed freely without removing this header.
*/

/*
	Compile: gcc -lptscan ptscan.c -o ptscan
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <pthread.h>
#include <signal.h>
#include <setjmp.h>


#define MAX_THREAD 1000

//funcs
int connectXT(int sock, struct sockaddr_in addr, int tout, int mode) ;
int writesoc(int sock, int wfd);
void sh(int signum);
void usage(char *argv);
void *doscan();
void in_read_timeout(int sig);


extern int 		errno;
unsigned long 		start, end; 
int     		port, 
			sltime = 1,
			sock,
			csleep = 5,
			serror, 
			isbackgr = 0,
			wfd = -1,
			readcount = 0, rsleep = 5,
			grpfound = 0, hidegrep = 0,
			lfd = -1,
			trapcount = 0;

char  			*logfile = NULL,
      			*grpstring = NULL, 
      			*wrdfile = NULL;
            
pthread_mutex_t 	start_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t 	stdout_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t 	logfile_mutex = PTHREAD_MUTEX_INITIALIZER;






//main
int main(int argc, char **argv)
{
 pid_t pipM;
 
 int    i = 1,
	args = 0,
	tc = 1;	
 
 char ch,
      *banner = "PortScanner v1.1 by Ralph, molnar_rcs@yahoo.com\n";     
 
 pthread_t thread_id[MAX_THREAD]; 
 //banner 
 fprintf(stderr, "%s", banner);
   
 //get options
 while ((ch = getopt(argc, argv, "t:b:e:s:p:T:l:w:r:R:g:BG")) != EOF)
 {
  	switch ((char) ch) 
	{
		case 't':
		{
 			// set proc count to 1st arg
 			tc = atoi(optarg);
			if(tc > MAX_THREAD)
			{
				fprintf(stderr, "Thred value #%d over the thread limit #%d.\n"
						"Edit MAX_THREAD and recompile if you want to use higher values.\n"
						"Aborting....\n", tc, MAX_THREAD);
				exit(0);
			}
			
			//create at least 1 scan thread
			if(tc < 1) tc = 1;
			//args++;			
 			break;
		}
		
		case 'b':
		{
 			start = ntohl(inet_addr(optarg));
			args += 1;			
 			break;
		}
		
		case 'e':
		{
			end = ntohl(inet_addr(optarg));
			args += 2;
			break;
		}
		
		case 's':
		{
			sltime = atoi(optarg);
   			if(sltime < 0)
    			{
				fprintf(stderr, "Sleep Time can't be negative\nAborting...\n");
				exit(0);
			}				
			break;
		}
		
		case 'T':
		{
			csleep = atoi(optarg);
   			if(csleep < 0)
    			{
				fprintf(stderr, "Connection Timeout can't be negative\nAborting...\n");
				exit(0);
			}
			
			if(csleep == 0)
			{
				fprintf(stderr, "Connection Timeout set to 0 is not adviced.\n");
				fprintf(stderr, "This usually generates improper scan results.\n");
			}				
			break;
		}
		
		case 'p':
		{
			port = atoi(optarg);
 			if(port > 65535 || port < 1)
  			{
				fprintf(stderr, "Port value most be between 1-65535.\n"
						"Aborting....\n");
				exit(0);
 			}			
			args += 4;
			break;
		}
		
		case 'l':
		{
			logfile = optarg;
			break;
		}
		
		case 'B':
		{
			isbackgr = 1;
			break;
		}
		
		case 'w':
		{
			wrdfile = optarg;			
			break;
		}
		
		case 'r':
		{
			readcount = atoi(optarg);
			break;
		}
		
		case 'R':
		{
			rsleep = atoi(optarg);
   			if(rsleep < 0)
    			{
				fprintf(stderr, "READ Timeout can't be negative\nAborting...\n");
				exit(0);
			}
			
			if(rsleep == 0)
			{
				fprintf(stderr, "READ Timeout set to 0 is not adviced.\n");
				fprintf(stderr, "This usually generates improper scan results.\n");
			}	
			break;
		}
		
		case 'g':
		{
			grpstring = optarg;
			break;
		}
		
		case 'G':
		{
			hidegrep = 1;
			break;
		}		
		
	}
 }
  
 if(args < 7)
 { 
   usage(argv[0]);
   exit(0);
 }

 // we don't want to let this happen
 if(start > end)
 {
  fprintf(stderr, "IP error. Start IP is greater than end IP.\nAborting...");
  exit(0);
 }
 
  
 
  if(logfile != NULL)
  {
   	fprintf(stderr, "Creating logfile %s.\n", logfile);      
 	lfd = open(logfile, O_WRONLY | O_CREAT | O_TRUNC, 0644); 
	if (lfd == -1)
	{
		perror(logfile);
		fprintf(stderr, "Could not open log file, logging is DISABLED.\n");
	}
	else write(lfd, banner, strlen(banner));
  }
  
  // is logging disabled and in background mode, exit
  if(isbackgr == 1 && lfd == -1)
   {
 	 fprintf(stderr, "Background mode: no logging, scan ABORTED.\n");
	 exit(0);
   }      
  
 
 // check for write to socket first 
 if(wrdfile != NULL)
  {
   	fprintf(stderr, "Opening data file %s\n", wrdfile);      
 	wfd = open(wrdfile, O_RDONLY); 
	if (wfd < 0)
	{
		perror(wrdfile);
		fprintf(stderr, "Writing to socket DISABLED.\n");
	}
  }
 
 // check for read from sock and grep
 if(grpstring != NULL && readcount == 0) 
 {
 	fprintf(stderr, "Read (-r) not enabled or `nr` is 0.\n"
			"Grep for string DISABLED\n");
	// silently disable -G here, actualy the code make no diffrence
	// since grep is disabled, no attemp will be made to log the grepstring
	// but just to be sure, let's set it to 0 
	hidegrep = 0;
	grpstring = NULL;	
 }
  
 //make sure threads don't mess up screen output
 pthread_mutex_lock( &start_mutex );
  
 // if in background mode, we fork here
 if(isbackgr == 1)
 {
 	pipM = fork();
	if(pipM == -1)
	 {
	  	perror("# Background mode:  Fork error. Aborting...");
	  	exit(0);
	 }
	
	if(pipM != 0)
	 {
	 	fprintf(stderr, "Forked into the background with PID [%d].\n", pipM);
		exit(1); // parent exists silently
	 }	
 }

 fprintf(stderr, "[+]\tCreating %d threads\n", tc);
 // loop: create threads
 for(i = 0; i < tc; i++)
 	 pthread_create( &thread_id[i], NULL, &doscan, NULL );

 
 if(isbackgr != 1) fprintf(stderr, "[+]\tStarting scan\n");
 fflush(stderr);
 fflush(stdout);
 pthread_mutex_unlock( &start_mutex );

 // wait for threads to finish the scan
 for(i = 0; i < tc; i++)
      pthread_join( thread_id[i], NULL); 
 
 if(lfd != -1) close(lfd);
 if(wfd != -1) close(wfd);

 return 1; 
}


// thread function
void *doscan()
{
 struct sockaddr_in addr;
 int i, j, logDS = 512, totalread;
 char *msg, *logdata = NULL, *screendata = NULL;
 
  
 if(isbackgr != 1) 
 	{
		if( (screendata = malloc ( 512 ) ) != NULL) bzero(screendata, 512);
		else return;
	}
	
 if(lfd != -1) 
	{
		// if we grep for a string, we need 512 bytes + grepstring size for the logged message
		// if not we need at least + rcount bytes of memory... rcount is 0 if we din't want
		// to read from the socket
		if(grpstring != NULL) 
			logDS += strlen(grpstring);
		else 
			logDS += readcount;
		
		if( (logdata = malloc ( logDS ) ) != NULL) bzero(logdata, logDS);
		else return;
	}
 
 do 
  { 	
  
  	//bzero log/screen buffer
	if(screendata != NULL) bzero(screendata, 512);
	if(logdata != NULL) bzero(logdata, logDS);
	
  	// lock
	pthread_mutex_lock( &start_mutex );
  	start++;
	pthread_mutex_unlock( &start_mutex );
	

	if(start > end + 1) break;		
	// skip 0 and 255
	if( (( (start - 1) & 0xff) == 255) || (( (start - 1) & 0xff) == 0) ) continue;
	
	
	
	//unlock
	
	
	// do da scan	
	if ((sock=socket(AF_INET,SOCK_STREAM,0)) == -1)
	{
		// might break here to avoid further errors
		// let's assume this is only a temporary error only in this thread
		if(isbackgr != 1) perror("SockeT:");
		continue;
	}
		
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port); 	
	addr.sin_addr.s_addr =  htonl(start - 1);
	
	
	if(isbackgr != 1 && screendata != NULL) 
		sprintf(screendata, "%s port %d: ", inet_ntoa(addr.sin_addr), port);
	
	if(lfd != -1 && logdata != NULL)
		sprintf(logdata, "%s port %d: ", inet_ntoa(addr.sin_addr) ,port);
	
	
		
	if ( (serror = connectXT(sock, addr, csleep, 0) ) == 0)
	{
		if(lfd != -1 && logdata != NULL) sprintf(logdata, "%sOPEN", logdata);
		if(isbackgr != 1 && screendata != NULL) sprintf(screendata, "%sOPEN", screendata);
		
		// port is open check for read, write, grep
		
		//if write, write content of datafile to socket
		if(wfd != -1) 
		{
			i = writesoc(sock, wfd);			
			// log on the fly
			if(lfd != -1 && logdata != NULL) sprintf(logdata, "%s * WRITTEN (%d) bytes", logdata, i);
			if(isbackgr != 1 && screendata != NULL) sprintf(screendata, "%s * WRITTEN (%d) bytes", screendata, i);
		}	
		
		// ivoke connectXT in read mode, we can store return value serror cuz 
		// the current value of serror has no use from here on
		if(readcount < 1) goto noread;
		
		if((serror = connectXT(sock, addr, rsleep, 1)) == 1)
		{
			//if read, read `nr` bytes from sock , if <= 0 skip step
			char c;
			msg = malloc(readcount + 1);
			if(msg) bzero(msg, readcount + 1);
			if(grpstring != NULL)  grpfound = 0;
			i = 0;
			int tcount = 0;
			
			while(msg && ( j = read(sock, &c, 1)) && ++i < readcount)
			{
				// detect error here
				if(j == -1) break;
				sprintf(msg, "%s%c", msg, c);
				if( grpstring != NULL && !grpfound && strstr(msg, grpstring) ) grpfound = 1;
				
				/* 
				  the next part is ugly, we use select to see if there's any data
				  if not, we wait again, at most 5 times, this means 10 seconds total
				  this also means that we expect the data to arrive at most in 5 chunks
				*/
				waitread:
				if( connectXT(sock, addr, 2, 1) != 1 ) 
				{
					tcount++;
					if(tcount <= 5) goto waitread;
					else break;
				}
			}
			
			// logging, print to screen
			if(isbackgr != 1 && screendata != NULL) 
			{
				if(grpfound == 0 && grpstring != NULL) 
					sprintf(screendata, "%s * READ (%d) bytes * NOTFND", screendata, i);
				
				if(grpfound == 1  && grpstring != NULL)
					sprintf(screendata, "%s * READ (%d) bytes * FOUND", screendata, i);
				
				if(grpstring == NULL) 
					sprintf(screendata, "%s * READ (%d) bytes", screendata, i);	
			}
			
			
			if(lfd != -1 && logdata != NULL) 
			{ 
				if(grpfound == 0 && grpstring != NULL) 
					sprintf(logdata, "%s * READ (%d) bytes * NOTFND",logdata, i);
				
					
				if(grpstring == NULL) 
				{
					sprintf(logdata, "%s * READ (%d) bytes * DATA -> ",logdata, i);
					for(j = 0; j < strlen(msg); j++)
					{
						// lets just make a check to avoid overflow
						if( !(strlen(logdata) < logDS) ) break;
						if( (msg[j] != '\r') && (msg[j] != '\n'))
						{
							// log on the fly	
							if(lfd != -1 && logdata != NULL) 
								sprintf(logdata, "%s%c", logdata, msg[j]);
						}
					}
				}
				
				if(grpfound == 1)
				{ 
					sprintf(logdata, "%s * READ (%d) bytes * FOUND", logdata, i);
					if(!hidegrep)  
						sprintf(logdata, "%s %s", logdata, grpstring);
				}
			}
			
			if(msg) free(msg);
		}
		else 
		{
			if(serror == 0)
			{
				if(isbackgr != 1 && screendata != NULL) sprintf(screendata, "%s * READ timeout", screendata);
				if(lfd != -1 && logdata != NULL)  sprintf(logdata, "%s * READ timeout", logdata);
			}
			else
			{
				if(isbackgr != 1 && screendata != NULL) 
					sprintf(screendata, "%s * READ %s", screendata, strerror(errno));
				if(lfd != -1 && logdata != NULL)  
					sprintf(logdata, "%s * READ %s", logdata, strerror(errno));
			}
		}
		
		noread:
		// end of read, write, grep		
		// log/print a newline
		if(isbackgr != 1 && screendata != NULL) sprintf(screendata, "%s\n", screendata);
		if(lfd != -1 && logdata != NULL)  sprintf(logdata, "%s\n", logdata);
		close(sock);
	}
	// connect timed out
	else 
	{
		if(lfd != -1 && logdata != NULL)  sprintf(logdata,  "%s%s\n", logdata, strerror(serror));
		if(isbackgr != 1 && screendata != NULL) sprintf(screendata, "%s%s\n", screendata, strerror(serror));
	}
	
	// write out log :)
	if(lfd != -1 && logdata != NULL)
	{
		// the locking below might not be necessary
		pthread_mutex_lock( &logfile_mutex );	
		write(lfd, logdata, strlen(logdata));
		pthread_mutex_unlock( &logfile_mutex );
	}
		
	// print to screen
	// the locking below might not be necessary
	if(isbackgr != 1 && screendata != NULL)
	{
		pthread_mutex_lock( &stdout_mutex );		
		fprintf(stdout, "%s", screendata);
		fflush (stdout);
		pthread_mutex_unlock( &stdout_mutex );
	}	
	// time to sleep	
	sleep(sltime);

  } while (start <= end);
  

  if(logdata != NULL) free(logdata);
  if(screendata != NULL) free(screendata);
  //end scan 
}

// connect, wait read
int connectXT(int sock, struct sockaddr_in addr, int tout, int mode) 
{ 
  int res, valopt; 
  long arg; 
  fd_set myset; 
  struct timeval tv; 
  socklen_t lon; 

  
  if(mode == 1)
  {
	// Initialize the file descriptor set
	FD_ZERO (&myset);
	FD_SET (sock, &myset);
	
	// Initialize the timeout data structure
	tv.tv_sec = tout;
	tv.tv_usec = 0;
	
	// select returns 0 if timeout, 1 if input available, -1 if error.
	return (select (FD_SETSIZE, &myset, NULL, NULL, &tv));
  }
  	
  // Set non-blocking 
  arg = fcntl(sock, F_GETFL, NULL); 
  arg |= O_NONBLOCK; 
  fcntl(sock, F_SETFL, arg); 

  // Trying to connect with timeout 
  res = connect(sock, (struct sockaddr *)&addr, sizeof(addr)); 

  if (res < 0) 
  { 
	if (errno == EINPROGRESS) 
	{ 
		tv.tv_sec = tout; 
		tv.tv_usec = 0; 
		FD_ZERO(&myset); 
		FD_SET(sock, &myset); 
		if (select(sock+1, NULL, &myset, NULL, &tv) > 0) 
		{ 
			lon = sizeof(int); 
			getsockopt(sock, SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon); 
			if (valopt) 
			{
				// quick hack to avoid valopt being out of the error range
				// if valopt is not a known error we assume that it's a time out error
				// actually it's not a good idea to do this here
				if(valopt < EPERM || valopt > EMEDIUMTYPE) valopt = 110;
				return valopt; 
			}
		} 
		else 
		{
			// quick hack to avoid valopt being out of the error range
			// if valopt is not a known error we assume that it's a time out error
			if(valopt < EPERM || valopt > EMEDIUMTYPE) valopt = 110;
			return valopt;
		}
	} 
	else return errno; 
  } 
  arg = fcntl(sock, F_GETFL, NULL); 
  arg &= (~O_NONBLOCK); 
  fcntl(sock, F_SETFL, arg); 
  return 0;
} 




// write to socket func
int writesoc(int sock, int wfd)
{
 int i = 0;
 char ch;
 
 //rewind file
 lseek(wfd, 0,SEEK_SET); 
 
 // read file byte by byte
 while(read(wfd, &ch, 1))
 {

	 // write byte to sock
	 if(write(sock, &ch, 1) != 1)
	 {
		// error occured or no byte was written
		// assume that this is a socket error and return from the func
		return i;
	 }
	 i++;
 }
 return i;
 
}

//usage
void usage(char *argv)
{
 fprintf(stderr, "[ARGS] \n"
	" -b begin         : start IP in dotted format - (x.x.x.x)\n"
	" -e end           : end IP in dotted format - (x.x.x.x)\n"
	" -p port          : port to scan\n"
	" [ -t threads ]   : number of threads to run, max %d - default is 1\n"
	" [ -s sleep ]     : sleep time in seconds for each thread - default is 1\n"
	" [ -T C_timeout ] : timeout in seconds before aborting a connection attempt - default value is 5\n"
	" [ -l logFile ]   : results will be also logged to logFile (this doesn't disable printing to the screen)\n"
	" [ -w file ]      : file that contains data that you want to write to the socket\n"
	"                    file can be binary or text\n"
	" [ -r nr ]        : reads at most `nr` bytes from the socket, can be used without [-w]\n"
	" [ -R R_timeout ] : timeout in seconds before aborting a read attempt - default value is 5\n"
	"                    works only with [-r nr] or else it's silently ignored\n"
	" [ -g string ]    : grep data read from socket for <string>\n"
	"                    can be used only with [-r]\n"
	"                    if <string> is found, `FOUND` will be displayed and `FOUND <string>` will be logged`\n"
	"                    if <string> is not found, NOTFND will be logged and/or displayed\n"	
	" [ -B ]           : run in background, printing to the screen will be disabled (errors, etc)\n"
	"                    works only with [-l logFile]\n"
	" [ -G ]           : specify this if you don't want <string> to be logged the log file\n"
	"                    useful when using binary data as <string> and you don't want the logfile messed up\n\n"
	" Read the README file if any for more information.\n\n"
	
	, MAX_THREAD);
 exit(0);
}


